/*$
Copyright (c) 2015, Azel
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the <organization> nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
$*/

/******************************
 * main
 ******************************/

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>

#include "mDef.h"

#include "zipfile.h"
#include "apperr.h"
#include "option.h"


//-----------------

OptionData g_opt;

#define VERSION_TEXT  "lhttpzip ver 1.0.0"

//-----------------

typedef struct
{
	int code;
	const char *mes;
}ErrMessage;

static const ErrMessage g_errmes[] = {
	{APPERR_MEMORY, "memory is not enough"},
	{APPERR_ZIP_OPENFILE, "can't open zip file"},
	{APPERR_ZIP_INVALID, "zip file is invalid"},
	{APPERR_ZIP_UNSUPPORTED, "unsupported zip format"},
	{APPERR_ZIP_NOZIP, "file is not zip"},
	{-1, NULL}
};

//-----------------
/* server.c */

void httpserv_close();
int httpserv_start(int port);
void httpserv_loop();

//-----------------



/** 終了処理 */

static void _finish()
{
	httpserv_close();
	zip_close();
}

/** エラー終了 */

static void _exit_err(int code)
{
	const ErrMessage *p;

	if(code != APPERR_SOCKET)
	{
		for(p = g_errmes; p->mes; p++)
		{
			if(code == p->code) break;
		}

		if(p->mes)
			fprintf(stderr, "error: %s\n", p->mes);
		else
			fputs("error\n", stderr);
	}

	_finish();
	exit(1);
}

/** ヘルプを表示して終了 */

static void _put_help(char *name)
{
	printf("Usage: %s [options] <zip file>\n\n", name);
	puts(
"Options:\n"
"   -p, --port <num>      port number (default: 8080)\n"
#ifdef HAVE_ICONV_H
"   -c, --charset <name>  charset of file name (default: UTF-8)\n"
"                         show list by 'iconv --list'\n"
#endif
"   -I, --no-index        always display the file list if uri is directory\n"
"                         (do not search index.html or index.htm)\n"
"   -n, --no-cache        do not cache the file in browser\n"
"   -v, --verbose         print HTTP headers\n"
"   -h, --help            show this help message\n"
"   -V, --version         show the version"
	);

	exit(0);
}

/** オプション処理
 *
 * @return ZIP ファイル名の位置 */

static int _get_option(int argc,char **argv)
{
	int n;
	struct option opts[] = {
		{"port", 1, 0, 'p'},
		{"charset", 1,0, 'c'},
		{"no-index", 0,0, 'I'},
		{"no-cache", 0, 0, 'n'},
		{"verbose", 0, 0, 'v'},
		{"help", 0,0, 'h'},
		{"version", 0,0, 'V'},
		{0,0,0,0}
	};

	while(1)
	{
		n = getopt_long(argc, argv, "p:c:IvnhV", opts, NULL);
		if(n < 0) break;

		switch(n)
		{
			case 'p':
				g_opt.port = atoi(optarg);
				break;
			case 'c':
				g_opt.charset = optarg;
				break;
			case 'I':
				g_opt.flags |= OPTION_F_NO_INDEX_HTML;
				break;
			case 'v':
				g_opt.flags |= OPTION_F_OUTPUT_HTTP_HEADER;
				break;
			case 'n':
				g_opt.flags |= OPTION_F_NO_CACHE;
				break;
			case 'h':
				_put_help(argv[0]);
				break;
			//バージョン情報
			case 'V':
				puts(VERSION_TEXT);
				exit(0);
				break;
		}
	}

	//ZIP ファイル指定がない

	if(optind == argc)
		_put_help(argv[0]);

	return optind;
}

/** 初期化 */

static int _init(int argc,char **argv)
{
	int ret,ind;

	//オプション初期値

	g_opt.port = 8080;
	g_opt.flags = 0;
	g_opt.charset = NULL;

	//オプション処理

	ind = _get_option(argc, argv);

	//ZIP 初期化

	if((ret = zip_init()))
		_exit_err(ret);

	//ZIP ファイル開く

	if((ret = zip_open(argv[ind])))
		_exit_err(ret);

	//HTTP サーバ起動
	
	if((ret = httpserv_start(g_opt.port)))
		_exit_err(ret);

	printf("* server start\nlook at http://localhost:%d/\n", g_opt.port);
	fflush(stdout);

	return 0;
}

/** メイン */

int main(int argc,char **argv)
{
	//初期化

	_init(argc, argv);

	//HTTP サーバループ

	httpserv_loop();

	//終了

	puts("\n* server stop");

	_finish();

	return 0;
}
