/*$
Copyright (c) 2015, Azel
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the <organization> nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
$*/

/************************************
 * HTTP リクエストパーサー
 ************************************/

#include <string.h>

#include "mDef.h"
#include "mStr.h"

#include "parser.h"
#include "apperr.h"


//------------------------

typedef struct
{
	int code;
	const char *reason;
}StatusValue;

static const StatusValue g_status[] = {
 {200, "OK"},
 {304, "Not Modified"},
 {400, "Bad Request"},
 {404, "Not Found"},
 {500, "Internal Server Error"},
 {501, "Not Implemented"},
 {0, NULL}
};

//------------------------

static char *g_bufend;   //バッファの終端ポインタ

//------------------------


/** req に HTTP ステータスコードセット */

void set_http_status_code(HTTPRequest *req,int code)
{
	const StatusValue *p;

	for(p = g_status; p->code; p++)
	{
		if(p->code == code)
		{
			req->statuscode = code;
			req->status_reason = p->reason;
		}
	}
}


//==========================
// 文字列操作
//==========================


/** 改行を NULL 文字に変換 */

static void _replace_enter_to_null(HTTPBuf *buf)
{
	char *pc;

	for(pc = buf->buf; *pc; pc++)
	{
		if(*pc == '\r' || *pc == '\n')
			*pc = 0;
	}
}

/** 次の行末まで移動 */

static char *_to_end(char *pc)
{
	return strchr(pc, 0);
}

/** 同じ行の次の空白または行末まで移動 */

static char *_to_space_or_end(char *pc)
{
	for(; *pc && *pc != ' '; pc++);

	return pc;
}

/** 行末から次の行へ移動 */

static char *_end_to_next_line(char *pc)
{
	for(; *pc == 0 && pc < g_bufend; pc++);

	return pc;
}


//==========================
// ヘッダ解析
//==========================


/** 次のヘッダ名と値を取得
 *
 * @return 次のヘッダ位置。NULL でエラー (ステータスコードもセットされる) */

static char *_get_next_header(char *pc,mStr *strname,mStr *strval,HTTPRequest *req)
{
	char *end,*nextl,*pc2;
	int firstline = 1;

	mStrEmpty(strval);

	while(*pc)
	{
		//改行位置と次行

		end = _to_end(pc);
		nextl = _end_to_next_line(end);

		//------ ヘッダ名取得

		if(firstline)
		{
			for(pc2 = pc; *pc2 && *pc2 != ':'; pc2++);

			if(*pc2 == 0)
			{
				set_http_status_code(req, HTTP_STATUSCODE_BAD_REQUEST);
				return NULL;
			}

			mStrSetTextLen(strname, pc, pc2 - pc);
			mStrLower(strname);

			firstline = 0;
			pc = pc2 + 1;
		}

		//------ 値

		//空白とタブをスキップ

		for(; *pc == ' ' || *pc == '\t'; pc++);

		mStrAppendText(strval, pc);

		//----- 次へ (次行が空白かタブで始まる場合、複数行で続く)

		pc = nextl;
		if(!(*pc == ' ' || *pc == '\t')) break;
	}

	return pc;
}


//==========================
// 解析 メイン
//==========================


/** 一行目を解析
 *
 * @return 二行目のポインタ。NULL でエラー。 */

static char *_parse_first_line(HTTPBuf *buf,HTTPRequest *req)
{
	char *pc,*pc2,*pcret;

	pc = buf->buf;

	//------ メソッド (GET/HEAD/POST のみ対応)

	pc2 = _to_space_or_end(pc);

	if(*pc2 == 0)
	{
		set_http_status_code(req, HTTP_STATUSCODE_BAD_REQUEST);
		return NULL;
	}

	if(strncmp(pc, "GET", 3) == 0)
		req->method = HTTP_METHOD_GET;
	else if(strncmp(pc, "HEAD", 4) == 0)
		req->method = HTTP_METHOD_HEAD;
	else if(strncmp(pc, "POST", 4) == 0)
		req->method = HTTP_METHOD_POST;
	else
	{
		//対応していないメソッド
		
		set_http_status_code(req, HTTP_STATUSCODE_NOT_IMPLEMENTED);
		return NULL;
	}

	pc = pc2 + 1;

	//------ リクエスト URI

	pc2 = _to_space_or_end(pc);

	if(*pc != '/' || *pc2 == 0)
	{
		set_http_status_code(req, HTTP_STATUSCODE_BAD_REQUEST);
		return NULL;
	}

	*pc2 = 0;

	req->path = pc + 1;  //最初の '/' は除く

	//? がある場合はパラメータ

	for(; *pc && *pc != '?'; pc++);

	if(*pc == '?' && pc[1])
	{
		*pc = 0;
		req->url_param = pc + 1;
	}
	else
		req->url_param = NULL;

	pc = pc2 + 1;

	//------ HTTP バージョン

	pc2 = _to_end(pc);
	pcret = _end_to_next_line(pc2);

	if(strcmp(pc, "HTTP/1.1") == 0)
		req->version = HTTP_VERSION_11;
	else if(strcmp(pc, "HTTP/1.0") == 0)
		req->version = HTTP_VERSION_10;
	else
		req->version = HTTP_VERSION_UNKNOWN;

	return pcret;
}

/** ヘッダを解析
 *
 * @return 0 で成功 */

static int _parse_headers(char *pc,HTTPRequest *req,char *modified_time)
{
	mStr name = MSTR_INIT, val = MSTR_INIT;
	char *pname;
	int ret = 1;

	while(*pc)
	{
		pc = _get_next_header(pc, &name, &val, req);
		if(!pc) goto ERR;

		pname = name.buf;

		if(strcmp(pname, "connection") == 0)
		{
			//Connection: (接続を継続するか)
			/* HTTP/1.0 では常に閉じる。
			 * HTTP/1.1 ではデフォルトで継続。明示的に close が指定された時のみ閉じる。 */

			if(strcasecmp(val.buf, "close") == 0)
				req->connection_close = 1;
		}
		else if(strcmp(pname, "if-modified-since") == 0)
		{
			/* If-Modified-Since: (指定日時以降に更新されているか)
			 * 更新されていなければ 304 を返す。 */

			/* ここでは、ZIP を開いた日時をファイルの更新時間とし、文字列でそのまま比較。
			 * 
			 * 前回開いた ZIP ファイルと新しく開いた ZIP ファイルで
			 * 同名のファイルがある場合があるので、
			 * 前の ZIP 内のファイルがキャッシュから取得されないようにするため。 */

			if(strcasecmp(val.buf, modified_time) == 0)
			{
				set_http_status_code(req, HTTP_STATUSCODE_NOT_MODIFIED);
				goto ERR;
			}
		}
	}

	ret = 0;

ERR:
	mStrFree(&name);
	mStrFree(&val);

	return ret;
}


//====================


/** HTTP リクエストを解析
 *
 * @param modified_time If-Modified-Sinc 用の比較文字列 */

void parser_http_request(HTTPBuf *buf,HTTPRequest *req,char *modified_time)
{
	char *pc;

	g_bufend = buf->bodytop - 4;

	//すべての改行を NULL 文字に

	_replace_enter_to_null(buf);

	//一行目解析

	pc = _parse_first_line(buf, req);
	if(!pc) return;

	//デフォルト値

	req->connection_close = (req->version == HTTP_VERSION_11)? 0: 1;

	//ヘッダ解析

	if(_parse_headers(pc, req, modified_time))
		return;

	//正常に受理

	set_http_status_code(req, 200);
}
