/*$
Copyright (c) 2015, Azel
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the <organization> nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
$*/

/************************************
 * HTTP サーバ
 ************************************/
/*
 * - SIGINT,SIGKILL シグナル受信でサーバ終了。
 */

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <errno.h>

#include "mDef.h"
#include "mList.h"
#include "mStr.h"

#include "parser.h"
#include "zipfile.h"
#include "apperr.h"
#include "option.h"


//--------------------
/** 接続中のクライアントのデータ */

typedef struct
{
	mListItem i;
	int sock;
}ClientItem;

//--------------------
/** サーバデータ */

typedef struct
{
	mList clist;     //クライアントのリスト
	int sock_serv,   //サーバのソケット
		fdcnt;       //select() 時の監視ファイルディスクリプタ数
	fd_set fdset;    //select() 時の監視対象
}ServerData;

static ServerData *g_serv = NULL;

//--------------------

#define _SOCKET_ERR  -1

#define _CITEM(p)        ((ClientItem *)(p))
#define _CITEM_TOP       ((ClientItem *)g_serv->clist.top)
#define _CITEM_NEXT(p)   ((ClientItem *)((p)->i.next))

//--------------------
/* html.c */

void html_directory_filelist(mStr *str,ZipFileInfo *info,const char *path_decode);

//--------------------


//============================
// sub
//============================


/** シグナルハンドラ */

void _signal_handle(int sig)
{

}

/** シグナル設定 */

static void _set_signal()
{
	struct sigaction sa;
	int i,sig[2] = {SIGINT, SIGKILL};

	memset(&sa, 0, sizeof(sa));
	sa.sa_handler = _signal_handle;

	for(i = 0; i < 2; i++)
		sigaction(sig[i], &sa, NULL);
}


//============================
// クライアント関連
//============================


/** クライアント数が増減した時の処理
 *
 * select() 用の値をセット。 */

static void _set_client_updown()
{
	ClientItem *pi;
	int max;
	fd_set fd;

	FD_ZERO(&fd);
	FD_SET(g_serv->sock_serv, &fd);

	max = g_serv->sock_serv;

	for(pi = _CITEM_TOP; pi; pi = _CITEM_NEXT(pi))
	{
		FD_SET(pi->sock, &fd);
	
		if(pi->sock > max)
			max = pi->sock;
	}

	g_serv->fdset = fd;
	g_serv->fdcnt = max + 1;
}

/** select() から戻った時、対象のクライアントを取得 */

static ClientItem *_get_client_from_fdset(fd_set *fds)
{
	ClientItem *pi;

	for(pi = _CITEM_TOP; pi; pi = _CITEM_NEXT(pi))
	{
		if(FD_ISSET(pi->sock, fds))
			return pi;
	}

	return NULL;
}

/** クライアントアイテム破棄時 */

static void _clientitem_destroy(mListItem *p)
{
	close(_CITEM(p)->sock);
}

/** クライアント追加 */

static int _add_client(int sock)
{
	ClientItem *pi;

	pi = (ClientItem *)mListAppendNew(&g_serv->clist, sizeof(ClientItem), _clientitem_destroy);
	if(!pi) return APPERR_MEMORY;

	pi->sock = sock;

	_set_client_updown();

	return APPERR_OK;
}

/** クライアント削除 */

static void _delete_client(ClientItem *pi)
{
	mListDelete(&g_serv->clist, M_LISTITEM(pi));

	_set_client_updown();
}


//============================
// sub - 送受信
//============================


/** リクエスト受信 */

static int _recv_request(ClientItem *pi,HTTPBuf *data)
{
	char *buf,*ptmp;
	int ret;
	ssize_t rret;
	uint32_t pos = 0,bufsize = 8192;

	//初期確保

	buf = (char *)mMalloc(bufsize, FALSE);
	if(!buf) return APPERR_MEMORY;

	//受信

	while(1)
	{
		rret = recv(pi->sock, buf + pos, bufsize - pos - 1, 0);

		if(rret == 0)
		{
			//切断された (クライアントを削除)

			_delete_client(pi);

			ret = APPERR_DISCONNECT;
			goto ERR;
		}
		else if(rret == -1)
		{
			//エラー

			ret = (errno == EINTR)? APPERR_QUIT: APPERR_SOCKET;
			goto ERR;
		}
		else
		{
			pos += rret;

			//ヘッダ部分を最後まで読み込んだか

			buf[pos] = 0;

			ptmp = buf + ((pos == rret || pos < 4)? 0: pos - rret - 4);

			ptmp = strstr(ptmp, "\r\n\r\n");
			if(ptmp) break;
			
			//バッファが一杯の場合、拡張

			if(pos == bufsize - 1)
			{
				bufsize += 8192;

				ptmp = (char *)mRealloc(buf, bufsize);
				if(!ptmp)
				{
					ret = APPERR_MEMORY;
					goto ERR;
				}

				buf = ptmp;
			}
		}
	}

	//正常終了

	*ptmp = 0;

	data->buf = buf;
	data->size = pos;
	data->bodytop = ptmp + 4;

	return APPERR_OK;

ERR:
	mFree(buf);
	return ret;
}

/** データ送信 */

static int _send_data(ClientItem *pi,void *buf,int size)
{
	ssize_t ret;
	uint8_t *ps = (uint8_t *)buf;

	while(size)
	{
		ret = send(pi->sock, ps, size, 0);

		if(ret == -1)
		{
			//エラー

			if(errno == EINTR)
				return APPERR_QUIT;
			else
				return APPERR_SOCKET;
		}
		else if(ret == 0)
		{
			_delete_client(pi);
			return APPERR_DISCONNECT;
		}
		else
		{
			ps += ret;
			size -= ret;
		}
	}

	return APPERR_OK;
}


//============================
// レスポンス sub
//============================


/** 標準出力に HTTP ヘッダ出力 */

static void _output_http_header(mStr *str)
{
	if(g_opt.flags & OPTION_F_OUTPUT_HTTP_HEADER)
	{
		fwrite(str->buf, 1, str->len, stdout);
		puts("-----------");
	}
}

/** 日時ヘッダを追加 */

static void _append_time_header(mStr *str,const char *headername,mBool bZipTime)
{
	struct tm *ptm;
	time_t t;
	char m[32];

	if(bZipTime)
		mStrAppendFormat(str, "%s: %s\r\n", headername, zip_get_opentime());
	else
	{
		//現在時刻
		
		time(&t);
	
		ptm = gmtime(&t);
		if(ptm)
		{
			strftime(m, 32, "%a, %d %b %Y %T GMT", ptm);
			mStrAppendFormat(str, "%s: %s\r\n", headername, m);
		}
	}
}

/** キャッシュ、更新日時ヘッダの追加 */

static void _append_cache_modified_header(mStr *str,HTTPRequest *req)
{
	if(g_opt.flags & OPTION_F_NO_CACHE)
	{
		//キャッシュしない => 更新日時は指定しない
		
		if(req->version == HTTP_VERSION_10)
			mStrAppendText(str, "Pragma: no-cache\r\n");
		else
			mStrAppendText(str, "Cache-Control: no-cache\r\n");
	}
	else
	{
		//キャッシュする => 更新日時指定。常に確認させる

		if(req->version != HTTP_VERSION_10)
			mStrAppendText(str, "Cache-Control: max-ape=0\r\n");

		_append_time_header(str, "Last-Modified", TRUE);
	}
}

/** HTTP レスポンスの基本形をセット */

static void _set_response_basic(mStr *str,HTTPRequest *req)
{
	//一行目

	mStrSetFormat(str, "HTTP/%s %d %s\r\n",
		(req->version == HTTP_VERSION_10)? "1.0": "1.1",
		req->statuscode, req->status_reason);

	//Date: 現在の時刻

	_append_time_header(str, "Date", FALSE);

	//Connection: 接続の継続 (HTTP/1.0 は非対応)

	if(req->version != HTTP_VERSION_10)
		mStrAppendFormat(str, "Connection: %s\r\n", (req->connection_close)? "close": "keep-alive");
}

/** ファイルデータを返す際のヘッダ部分をセット */

static void _set_response_filehead(mStr *str,ZipFileInfo *info,HTTPRequest *req)
{
	_set_response_basic(str, req);

	mStrAppendFormat(str, "Content-Length: %d\r\n", info->size);

	_append_cache_modified_header(str, req);

	_output_http_header(str);

	mStrAppendText(str, "\r\n");
}


//============================
// レスポンス 送信
//============================


/** HTTP レスポンス (エラー) を送信
 *
 * @param code  ステータスコード。0 で req の値を使う
 * @param message NULL でなし */

static int _send_response_error(ClientItem *pi,int code,const char *message,HTTPRequest *req)
{
	mStr str = MSTR_INIT,strbody = MSTR_INIT;
	int ret;

	if(code)
		set_http_status_code(req, code);

	//ボディ部分

	mStrSetFormat(&strbody, "[%d %s]", req->statuscode, req->status_reason);

	if(message)
		mStrAppendFormat(&strbody, "<br><br>%s", message);

	//ヘッダ + body

	_set_response_basic(&str, req);

	mStrAppendText(&str, "Content-Type: text/html; charset=UTF-8\r\n");
	mStrAppendFormat(&str, "Content-Length: %d\r\n", strbody.len);

	_output_http_header(&str);

	mStrAppendText(&str, "\r\n");
	mStrAppendStr(&str, &strbody);

	//送信

	ret = _send_data(pi, str.buf, str.len);

	mStrFree(&str);
	mStrFree(&strbody);

	return ret;
}

/** 生成した HTML を送信
 *
 * info,path が NULL でトップページ */

static int _send_response_html(ClientItem *pi,
	ZipFileInfo *info,const char *path,HTTPRequest *req)
{
	mStr str,html;
	int ret;

	//最初に文字列バッファ確保

	mStrAlloc(&str, 4096);
	mStrAlloc(&html, 4096);

	//html 内容

	html_directory_filelist(&html, info, path);

	//レスポンス

	_set_response_basic(&str, req);

	mStrAppendFormat(&str, "Content-Type: text/html; charset=UTF-8\r\n");
	mStrAppendFormat(&str, "Content-Length: %d\r\n", html.len);
	_append_cache_modified_header(&str, req);

	_output_http_header(&str);

	mStrAppendText(&str, "\r\n");

	if(req->method == HTTP_METHOD_GET)
		mStrAppendStr(&str, &html);

	mStrFree(&html);

	//送信

	ret = _send_data(pi, str.buf, str.len);

	mStrFree(&str);

	return ret;
}


//============================
// リクエスト/レスポンス処理
//============================


/** パスとファイル名を結合 */

static void _path_combine(mStr *strdst,mStr *path,const char *filename)
{
	if(mStrIsEmpty(path))
		mStrSetText(strdst, filename);
	else
	{
		mStrSetText(strdst, path->buf);
		mStrAppendText(strdst, filename);
	}
}

/** パスがディレクトリの場合、ファイルリストを表示するか
 *
 * @return TRUE でファイルリスト、FALSE で info のファイル */

static int _is_directory_filelist(mStr *strpath,ZipFileInfo *info)
{
	mStr str = MSTR_INIT;
	int filelist = FALSE;
	
	if(g_opt.flags & OPTION_F_NO_INDEX_HTML)
		//常にファイルリスト
		filelist = TRUE;
	else
	{
		//index.html or index.htm を探す (なければファイルリスト)

		_path_combine(&str, strpath, "index.html");

		if(zip_search(str.buf, info))
		{
			_path_combine(&str, strpath, "index.htm");
			
			if(zip_search(str.buf, info))
				filelist = TRUE;
		}

		mStrFree(&str);
	}

	return filelist;
}

/** レスポンス送信 (ZIP 内ファイル)
 *
 * @param str  パス。空文字列の場合はルート */

static int _send_response_inzip(ClientItem *pi,HTTPRequest *req,mStr *str)
{
	ZipFileInfo info;
	uint8_t *filebuf = NULL;
	int ret = APPERR_OK,is_root;

	is_root = mStrIsEmpty(str);
	
	//ファイル検索 (ルートの場合は除く)

	if(!is_root)
	{
		if(zip_search(str->buf, &info))
			return _send_response_error(pi, HTTP_STATUSCODE_NOT_FOUND, str->buf, req);

		/* URI が '/' で終わっていないのに対象がディレクトリの場合、
		 * ファイルとして扱うため、エラーにする */

		if((info.flags & ZIPFILEINFO_F_DIRECTORY) && mStrGetLastChar(str) != '/')
			return _send_response_error(pi, HTTP_STATUSCODE_NOT_FOUND, "file is directory", req);
	}

	//ディレクトリの場合 (ルートの場合は常にディレクトリ扱い)

	if(is_root || (info.flags & ZIPFILEINFO_F_DIRECTORY))
	{
		if(_is_directory_filelist(str, &info))
		{
			//ファイルリスト

			if(is_root)
				return _send_response_html(pi, NULL, NULL, req);
			else
				return _send_response_html(pi, &info, str->buf, req);
		}
	}

	//ファイルデータを取得 (HEAD の場合は除く)

	if(req->method == HTTP_METHOD_GET)
	{
		if(zip_decode(&info, &filebuf))
			return _send_response_error(pi, HTTP_STATUSCODE_INTERNAL_SERVER_ERROR, "failed to decode zip", req);
	}

	//レスポンスを送信

	_set_response_filehead(str, &info, req);

	ret = _send_data(pi, str->buf, str->len);

	if(ret == APPERR_OK && req->method == HTTP_METHOD_GET)
	{
		//ファイルデータ (HEAD の場合は送信しない)

		ret = _send_data(pi, filebuf, info.size);
	}

	//

	mFree(filebuf);

	return ret;
}

/** レスポンスを送信
 *
 * [!] エラー時も常にレスポンス送信すること */

static int _send_response_ok(ClientItem *pi,HTTPBuf *hbuf,HTTPRequest *req)
{
	mStr str = MSTR_INIT;
	int ret;

	//URL エンコードをデコード (UTF-8 とみなす)
	/* - path は先頭の '/' は除外されている。終端の '/' はある。
	 * - ルートの場合 str は空となる */

	mStrDecodePercentEncoding(&str, req->path);

	//各処理

	if(req->method == HTTP_METHOD_POST)
	{
		//[POST] (今は未対応)

		ret = _send_response_error(pi, HTTP_STATUSCODE_NOT_IMPLEMENTED, "POST method is not supported", req);
	}
	else
	{
		//[GET/HEAD] ZIP 内ファイル

		ret = _send_response_inzip(pi, req, &str);
	}

	mStrFree(&str);

	return ret;
}

/** リクエストとレスポンスの処理 */

static int _proc_request_and_response(fd_set *fdset)
{
	ClientItem *pi;
	HTTPBuf hbuf;
	HTTPRequest req;
	int ret;

	//クライアント取得

	pi = _get_client_from_fdset(fdset);
	if(!pi) return APPERR_OK;

	//リクエスト受信 => hbuf

	ret = _recv_request(pi, &hbuf);
	if(ret) return ret;

	//ヘッダ出力

	if(g_opt.flags & OPTION_F_OUTPUT_HTTP_HEADER)
	{
		puts((char *)hbuf.buf);
		puts("-----------");
	}

	//リクエスト解析
	/* req.statuscode == 200 で正常終了 */

	parser_http_request(&hbuf, &req, zip_get_opentime());

	//レスポンス送信

	if(req.statuscode != 200)
		ret = _send_response_error(pi, 0, NULL, &req);
	else
		ret = _send_response_ok(pi, &hbuf, &req);

	//接続を閉じる
	
	if(req.connection_close)
		_delete_client(pi);

	//

	mFree(hbuf.buf);

	return ret;
}


//============================
// main
//============================


/** HTTP サーバを閉じる */

void httpserv_close()
{
	if(g_serv)
	{
		//クライアント削除
		
		mListDeleteAll(&g_serv->clist);

		//ソケット閉じる

		if(g_serv->sock_serv != _SOCKET_ERR)
			close(g_serv->sock_serv);

		mFree(g_serv);
		g_serv = NULL;
	}
}

/** HTTP サーバを起動
 *
 * 失敗時はエラー表示。 */

int httpserv_start(int port)
{
	struct sockaddr_in addr;
	int sock,n;

	//シグナル設定

	_set_signal();

	//確保

	g_serv = (ServerData *)mMalloc(sizeof(ServerData), TRUE);
	if(!g_serv) return APPERR_MEMORY;

	//ソケット作成
	
	g_serv->sock_serv = sock = socket(AF_INET, SOCK_STREAM, 0);
	if(sock == _SOCKET_ERR) goto ERR;

	//ソケットオプションセット
	/* "Address already in use" のエラーが出るのを防ぐ */

	n = 1;
	setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &n, sizeof(int));

	//名前付け

	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	addr.sin_port = htons(port);

	if(bind(sock, (struct sockaddr *)&addr, sizeof(addr)))
		goto ERR;

	//listen

	if(listen(sock, 5))
		goto ERR;

	//select() 用の値

	FD_ZERO(&g_serv->fdset);
	FD_SET(g_serv->sock_serv, &g_serv->fdset);

	g_serv->fdcnt = g_serv->sock_serv + 1;

	return APPERR_OK;

	//---- socket error
ERR:
	perror("socket error");
	httpserv_close();
	
	return APPERR_SOCKET;
}

/** ループ */

void httpserv_loop()
{
	fd_set fds;
	int ret,sock;

	while(1)
	{
		fds = g_serv->fdset;

		ret = select(g_serv->fdcnt, &fds, NULL, NULL, NULL);

		if(ret == 0)
			continue;
		else if(ret < 0)
		{
			//エラー (シグナル割り込みでサーバ終了)

			if(errno == EINTR) break;
		}
		else if(FD_ISSET(g_serv->sock_serv, &fds))
		{
			//サーバ : クライアントからの接続要求

			sock = accept(g_serv->sock_serv, NULL, NULL);

			if(sock == _SOCKET_ERR)
				perror("socket error");
			else
			{
				//クライアント追加

				if(_add_client(sock))
					close(sock);
			}
		}
		else
		{
			//クライアント : リクエスト/レスポンス処理

			ret = _proc_request_and_response(&fds);

			if(ret == APPERR_QUIT)
				break;
			else if(ret == APPERR_SOCKET)
				perror("socket error");
		}
	}
}

