/*$
Copyright (c) 2015, Azel
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the <organization> nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
$*/

/***************************************
 * ZIP ファイル操作
 ***************************************/
/*
 * - 無圧縮と Deflate 圧縮のみ対応。
 * - ZIP 内のファイル名は UTF-8 と明示されていない限り、
 *   作成されたシステム上での文字エンコードとなるので、
 *   自動判別させないと文字エンコードを判定できない。
 * - オプションで文字エンコードが指定されていない場合は UTF-8 として扱う。
 */

#include "mConfig.h"

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <zlib.h>

#ifdef HAVE_ICONV_H
#include <iconv.h>
#include <errno.h>
#endif

#include "mDef.h"
#include "mTree.h"
#include "mStr.h"

#include "apperr.h"
#include "zipfile.h"
#include "option.h"


//--------------------
/** ファイルツリーアイテム */

typedef struct
{
	mTreeItem i;

	uint8_t method,    //圧縮メソッド (0 or 8)
		flags;
	int name_len;      //ファイル名長さ (NULL 文字除く)
	uint32_t compsize, //圧縮サイズ
		uncompsize,    //非圧縮サイズ
		hash;          //ファイル名のハッシュ値
	fpos_t offset;     //データのオフセット位置 (ファイル先頭から)
	char name[1];      //ファイル名
}FileTreeItem;

enum
{
	FILETREEITEM_F_DIR = 1<<0  //ディレクトリ
};

//--------------------
/** ZIP 操作用データ */

typedef struct
{
	mTree tree;  //ファイルツリー
	FILE *fp;

#ifdef HAVE_ICONV_H
	iconv_t iconv;
	mStr strtmp;
#endif

	char opentime[32];	//開いた時間の日時文字列 (GMT)

	z_stream z;			//zlib
	uint8_t *decodebuf; //デコード用バッファ
}ZipData;

static ZipData *g_zip = NULL;

//--------------------

#define _ICONV_ERROR    ((iconv_t)-1)
#define _ICONV_BUFSIZE  128

#define _METHOD_STORED  0
#define _METHOD_DEFLATE 8

#define _VAL16(p)  *((uint16_t *)(p))
#define _VAL32(p)  *((uint32_t *)(p))

#define _ITEM(p)       ((FileTreeItem *)(p))
#define _ITEM_NEXT(p)  ((FileTreeItem *)((p)->i.next))

#define _DECODE_BUFSIZE  (16 * 1024)

//--------------------



//==========================
// sub
//==========================


/** 読み込んだデータを litte -> システムエンディアンに変換 */

static void _convert_endian(uint8_t *buf,const char *format)
{
	for(; *format; format++)
	{
		if(*format == '2')
		{
			//2byte

			*((uint16_t *)buf) = (buf[1] << 8) | buf[0];
			buf += 2;
		}
		else
		{
			//4byte

			*((uint32_t *)buf) = ((uint32_t)buf[3] << 24) | (buf[2] << 16) | (buf[1] << 8) | buf[0];
			buf += 4;
		}
	}
}

/** ファイル名変換
 *
 * - 文字エンコード変換。
 * - パス区切りを '/' に統一。 */

static void _convert_filename(mStr *strdst,mBool utf8name)
{
	//文字エンコードを UTF-8 へ変換
	/* UTF-8 で格納されているならそのまま */

#ifdef HAVE_ICONV_H

	if(g_zip->iconv != _ICONV_ERROR && !utf8name)
	{
		char *ps,*pd,buf[_ICONV_BUFSIZE];
		size_t ssize,dsize,ret;

		ps = strdst->buf;
		pd = buf;
		ssize = strdst->len;
		dsize = _ICONV_BUFSIZE;

		//変換状態初期化

		iconv(g_zip->iconv, NULL, NULL, NULL, NULL);

		//strtmp に変換

		mStrEmpty(&g_zip->strtmp);

		while(ssize > 0)
		{
			ret = iconv(g_zip->iconv, &ps, &ssize, &pd, &dsize);

			mStrAppendTextLen(&g_zip->strtmp, buf, _ICONV_BUFSIZE - dsize);

			pd = buf;
			dsize = _ICONV_BUFSIZE;

			//無効な文字があったら変換なしとする
			/* (E2BIG の場合はバッファが足りない) */
			
			if(ret < 0 && errno != E2BIG) break;
		}

		//正常終了したらコピー

		if(ssize == 0)
			mStrCopy(strdst, &g_zip->strtmp);
	}
	
#endif

	//パス区切り

	mStrReplaceChar(strdst, '\\', '/');
}

/** 文字列ハッシュ値計算 */

static uint32_t _calc_hash(const char *pc,int len)
{
	uint32_t h = 0;

	for(; len && *pc; pc++, len--)
		h = h * 37 + (*pc);

	return h;
}

/** ZipFileInfo にファイル情報セット */

static void _set_zipfileinfo(ZipFileInfo *info,FileTreeItem *pi)
{
	info->name = pi->name;
	info->size = pi->uncompsize;

	info->flags = 0;
	if(pi->flags & FILETREEITEM_F_DIR) info->flags |= ZIPFILEINFO_F_DIRECTORY;
}


//==========================
// ツリー検索
//==========================


/** 指定階層のみからファイル検索 (子は辿らない)
 *
 * @param parent     NULL でルート
 * @param directory  [1] ディレクトリのみ [-1] ファイルタイプは問わない */

static FileTreeItem *_search_no_child(FileTreeItem *parent,const char *filename,int len,int directory)
{
	FileTreeItem *pi;
	uint32_t hash;

	hash = _calc_hash(filename, len);

	pi = (parent)? _ITEM(parent->i.first): _ITEM(g_zip->tree.top);

	for(; pi; pi = _ITEM_NEXT(pi))
	{
		if(pi->hash == hash
			&& len == pi->name_len
			&& strncmp(pi->name, filename, len) == 0)
		{
			//ファイル名が一致
			
			if(directory < 0)
				return pi;
			else
				//ファイルタイプが一致しないなら NULL
				return (!directory == !(pi->flags & FILETREEITEM_F_DIR))? pi: NULL;
		}
	}

	return NULL;
}

/** ツリー全体から指定パスを検索 */

static FileTreeItem *_search_all(const char *path)
{
	const char *top,*next;
	int len;
	FileTreeItem *parent = NULL,*pi;

	top = path;
	if(*top == 0) return NULL;

	while(1)
	{
		//次の '/' か終端

		for(next = top; *next && *next != '/'; next++);

		//現在の階層から検索

		len = next - top;

		pi = _search_no_child(parent, top, len, (*next == '/')? 1: -1);
		if(!pi) return NULL;

		//

		if(*next == 0 || next[1] == 0)
			//終了
			/* '/' で終わっている場合は現在のディレクトリを返す */
			return pi;
		else
		{
			//続ける

			parent = pi;
			top = next + 1;
		}
	}

	return NULL;
}


//==========================
// ファイルツリー sub
//==========================


/** ツリーアイテム作成 (ファイル/ディレクトリ共通)
 *
 * @param parent NULL でルート */

static FileTreeItem *_new_item(FileTreeItem *parent,
	const char *name,int len,mBool directory)
{
	FileTreeItem *pi,*ins;

	//作成

	pi = (FileTreeItem *)mTreeAppendNew(&g_zip->tree, M_TREEITEM(parent), sizeof(FileTreeItem) + len, NULL);
	if(!pi) return NULL;

	//ファイル名
	
	memcpy(pi->name, name, len);
	pi->name[len] = 0;
	pi->name_len = len;

	//ファイル名ハッシュ値

	pi->hash = _calc_hash(name, len);

	//ディレクトリ

	if(directory)
	{
		pi->flags |= FILETREEITEM_F_DIR;

		//ファイルより前の位置に置く

		ins = (FileTreeItem *)mTreeGetFirstItem(&g_zip->tree, M_TREEITEM(parent));

		for(; ins && (ins->flags & FILETREEITEM_F_DIR); ins = _ITEM_NEXT(ins));

		if(ins)
			mTreeMoveItem(&g_zip->tree, M_TREEITEM(pi), M_TREEITEM(ins));
	}

	return pi;
}

/** パスから、ファイルアイテムを作成するための親アイテムを取得し、パスのファイル名部分を返す
 *
 * 親ディレクトリが未作成なら、作成する。
 * 自身がディレクトリなら、作成して 0 を返す。
 *
 * @param ppname  [IN] パス [OUT] ディレクトリを除いたファイル名部分へのポインタが入る。
 * @return ファイル名長さ (0 でファイルアイテム作成しない、-1 で失敗) */

static int _get_new_parent(FileTreeItem **ppret,char **ppname)
{
	char *top,*next;
	FileTreeItem *parent = NULL,*pi;
	int len;

	top = *ppname;

	//ファイル名がない場合は何もしない (標準入力からのファイルなど)
	if(*top == 0) return 0;

	while(1)
	{
		//次の '/' か終端

		for(next = top; *next && *next != '/'; next++);

		//next が終端で、ファイル

		if(*next == 0)
		{
			*ppret = parent;
			*ppname = top;

			return next - top;
		}

		//ディレクトリ検索

		len = next - top;

		pi = _search_no_child(parent, top, len, TRUE);
		if(!pi)
		{
			//存在しない場合は、ディレクトリ作成

			pi = _new_item(parent, top, len, TRUE);
			if(!pi) return -1;
		}

		parent = pi;

		//'/' で終わっている場合はディレクトリなので、ファイルアイテムなし

		if(next[1] == 0) return 0;

		//次へ

		top = next + 1;
	}

	return 0;
}

/** ファイルパスからツリーアイテム作成
 *
 * @return [0] 成功 [1] 何もせず次へ [-1] 失敗 */

static int _new_treeitem(FileTreeItem **ppitem,char *filename)
{
	FileTreeItem *pi,*parent;
	int len;

	len = _get_new_parent(&parent, &filename);

	if(len < 0)
		return -1;
	else if(len == 0)
		return 1;
	else
	{
		//ファイル用アイテム作成
		
		pi = _new_item(parent, filename, len, FALSE);
		if(!pi)
			return -1;
		else
		{
			*ppitem = pi;
			return 0;
		}
	}
}


//================================
// ファイルツリー作成 (ZIP 解析)
//================================


/** 中央ディレクトリへ移動*/

static int _goto_central_directory(FILE *fp)
{
	uint8_t b[18];
	int i,flag = 0;

	//先頭が 'PK' でなければ ZIP ファイルでない

	if(fread(b, 1, 2, fp) != 2 || (b[0] != 'P' && b[1] != 'K'))
		return APPERR_ZIP_NOZIP;

	//識別子検索
	/* コメントが格納されている場合があるので、最大長さまで検索 */

	fseek(fp, -22, SEEK_END);

	for(i = 0xffff; i >= 0; i--)
	{
		if(fread(b, 1, 4, fp) != 4) break;

		if(b[0] == 0x50 && b[1] == 0x4b && b[2] == 0x05 && b[3] == 0x06)
		{
			flag = 1;
			break;
		}

		if(fseek(fp, -5, SEEK_CUR)) break;
	}

	if(!flag) return APPERR_ZIP_INVALID;

	//データ

	if(fread(b, 1, 18, fp) != 18) return APPERR_ZIP_INVALID;

	_convert_endian(b, "2222442");

	if(_VAL16(b) != 0) return APPERR_ZIP_UNSUPPORTED;

	//中央ディレクトリの位置

	fseek(fp, _VAL32(b + 12), SEEK_SET);

	return APPERR_OK;
}

/** ローカルファイルヘッダからファイルデータ位置へ */

static int _goto_filedata(FILE *fp)
{
	uint8_t h[30];

	if(fread(h, 1, 30, fp) != 30) return APPERR_ZIP_INVALID;

	_convert_endian(h, "42222244422");

	//識別子

	if(_VAL32(h) != 0x04034b50) return APPERR_ZIP_INVALID;

	//ファイル名と拡張フィールドをスキップ

	fseek(fp, _VAL16(h + 26) + _VAL16(h + 28), SEEK_CUR);
	
	return APPERR_OK;
}

/** ファイルツリー作成
 *
 * @return エラーコード */

static int _read_filelist(FILE *fp)
{
	uint8_t h[46];
	uint16_t flags,method,len_fname;
	uint32_t size,uncompsize;
	fpos_t fpos;
	FileTreeItem *pi;
	mStr strname;
	int ret = APPERR_ZIP_INVALID,n;

	//中央ディレクトリへ

	ret = _goto_central_directory(fp);
	if(ret) return ret;

	//ファイル名用文字列確保

	mStrAlloc(&strname, 1024);

	//

	while(1)
	{
		//読み込み (可変部分は除く)

		if(fread(h, 1, 46, fp) != 46)
			goto ERR;

		//エンディアン変換

		_convert_endian(h, "42222224442222244");

		//--------- データ取得

		//識別子

		if(_VAL32(h) != 0x02014b50) break;

		//

		flags  = _VAL16(h + 8);
		method = _VAL16(h + 10);

		if((_VAL16(h + 6) & 0xff) > 20	//展開に必要な最小バージョン (下位バイト)
			|| (flags & 1)				//汎用フラグ (暗号化は未対応)
			|| (method != _METHOD_STORED && method != _METHOD_DEFLATE))	//圧縮メソッド (無圧縮、Deflate 以外は未対応)
		{
			ret = APPERR_ZIP_UNSUPPORTED;
			goto ERR;
		}

		//圧縮サイズ、非圧縮サイズ、ファイル名長さ

		size = _VAL32(h + 20);
		uncompsize = _VAL32(h + 24);
		len_fname = _VAL16(h + 28);

		//ファイル名

		if(!mStrResize(&strname, len_fname))
		{
			ret = APPERR_MEMORY;
			goto ERR;
		}

		if(fread(strname.buf, 1, len_fname, fp) != len_fname)
			goto ERR;

		mStrSetLen(&strname, len_fname);

		_convert_filename(&strname, ((flags & (1<<11)) != 0));

		//拡張フィールド、コメントをスキップ

		fseek(fp, _VAL16(h + 30) + _VAL16(h + 32), SEEK_CUR);

		//------- ローカルファイルヘッダへ飛び、データ位置へ

		fgetpos(fp, &fpos);
		fseek(fp, _VAL32(h + 42), SEEK_SET);

		ret = _goto_filedata(fp);
		if(ret) goto ERR;
		
		//------- アイテム作成

		n = _new_treeitem(&pi, strname.buf);
		if(n < 0)
		{
			ret = APPERR_MEMORY;
			goto ERR;
		}
		else if(n == 0)
		{
			//ファイルアイテム
			
			fgetpos(fp, &pi->offset);

			pi->method = method;
			pi->compsize = size;
			pi->uncompsize = uncompsize;
		}

		fsetpos(fp, &fpos);
	}

	ret = APPERR_OK;

ERR:
	mStrFree(&strname);

	return ret;
}


//==========================
// デコード
//==========================


/** 無圧縮 デコード */

static int _decode_stored(FileTreeItem *pi,uint8_t *outbuf)
{
	size_t n;

	if(pi->compsize != pi->uncompsize)
		return APPERR_ZIP_INVALID;

	n = fread(outbuf, 1, pi->compsize, g_zip->fp);

	return ((uint32_t)n == pi->compsize)? APPERR_OK: APPERR_ZIP_INVALID;
}

/** Deflate デコード */

static int _decode_deflate(FileTreeItem *pi,uint8_t *outbuf)
{
	z_stream *pz;
	uint8_t *inbuf;
	uint32_t remain;
	int ret = APPERR_ZIP_DECODE,zret,size;

	pz = &g_zip->z;
	inbuf = g_zip->decodebuf;
	remain = pi->compsize;

	//zlib

	inflateReset(pz);

	pz->next_out = outbuf;
	pz->avail_out = pi->uncompsize;

	//

	while(remain)
	{
		//ファイルから読み込み
		
		if(pz->avail_in == 0)
		{
			size = fread(inbuf, 1, (remain < _DECODE_BUFSIZE)? remain: _DECODE_BUFSIZE, g_zip->fp);
			if(size == 0) break;

			pz->next_in = inbuf;
			pz->avail_in = size;
			
			remain -= size;
		}

		//展開

		zret = inflate(pz, Z_NO_FLUSH);

		if(zret == Z_STREAM_END)
		{
			ret = APPERR_OK;
			break;
		}
		else if(zret != Z_OK)
			break;
	}

	return ret;
}


//==========================
// main
//==========================
/*
 * - 戻り値はすべて APPERR_*
 */


/** 最初の初期化 */

int zip_init()
{
	//確保

	g_zip = (ZipData *)mMalloc(sizeof(ZipData), TRUE);
	if(!g_zip) return APPERR_MEMORY;

	//デコード用バッファ確保

	g_zip->decodebuf = (uint8_t *)mMalloc(_DECODE_BUFSIZE, FALSE);
	if(!g_zip->decodebuf)
		return APPERR_MEMORY;

	//zlib 初期化

	if(inflateInit2(&g_zip->z, -15) != Z_OK)
		return APPERR_UNKNOWN;

	return APPERR_OK;
}

/** ZIP ファイル閉じる */

void zip_close()
{
	if(g_zip)
	{
		//ファイルツリー
		
		mTreeDeleteAll(&g_zip->tree);

		//ファイル
		
		if(g_zip->fp)
			fclose(g_zip->fp);

		//zlib

		inflateEnd(&g_zip->z);

		mFree(g_zip->decodebuf);

		//

		mFree(g_zip);
		g_zip = NULL;
	}
}

/** ZIP ファイルを開く */

int zip_open(const char *filename)
{
	int ret;
	struct tm *ptm;
	time_t t;

	//開く

	g_zip->fp = fopen(filename, "rb");
	if(!g_zip->fp) return APPERR_ZIP_OPENFILE;

	//文字エンコード変換用初期化
	/* (指定なし、または失敗時は UTF-8 として扱う) */

#ifdef HAVE_ICONV_H

	g_zip->iconv = _ICONV_ERROR;

	if(g_opt.charset)
	{
		if((g_zip->iconv = iconv_open("UTF-8", g_opt.charset)) == _ICONV_ERROR)
		{
			//対応していない形式
			
			if(errno == EINVAL)
				fprintf(stderr, "! iconv unsupported charset : '%s'\n", g_opt.charset);
		}
	}

	if(g_zip->iconv != _ICONV_ERROR)
		mStrAlloc(&g_zip->strtmp, 1024);

#endif

	//ファイルリスト作成

	ret = _read_filelist(g_zip->fp);

	//文字エンコード変換、終了

#ifdef HAVE_ICONV_H

	if(g_zip->iconv != _ICONV_ERROR)
	{
		iconv_close(g_zip->iconv);
		mStrFree(&g_zip->strtmp);
	}
	
#endif

	if(ret) return ret;

	//# Debug: ファイルリスト表示

/*
	FileTreeItem *pi;

	for(pi = _ITEM(g_zip->tree.top); pi; pi = (FileTreeItem *)mTreeItemGetNext(M_TREEITEM(pi)))
	{
		printf("[%s](%d) :hash 0x%x:method %d:flags %x:size %u:uncomp %u\n",
			pi->name, pi->name_len, pi->hash, pi->method, pi->flags,
			pi->compsize, pi->uncompsize);
	}
*/

	//開いた日時の文字列 (HTTP レスポンス&日時比較用)

	time(&t);
	ptm = gmtime(&t);

	if(ptm)
		strftime(g_zip->opentime, 32, "%a, %d %b %Y %T GMT", ptm);

	return APPERR_OK;
}

/** 開いた時刻の文字列をポインタで取得 */

char *zip_get_opentime()
{
	return g_zip->opentime;
}

/** パスからファイル検索
 *
 * @param path  終端がパス区切りならディレクトリ */

int zip_search(const char *path,ZipFileInfo *info)
{
	FileTreeItem *pi;

	pi = _search_all(path);
	if(!pi) return APPERR_ZIP_NOTFIND;

	info->item = pi;

	_set_zipfileinfo(info, pi);

	return APPERR_OK;
}

/** info の位置を元にアイテムを移動
 *
 * @return 0 で終了 */

int zip_move(ZipFileInfo *info,int type)
{
	FileTreeItem *pi = (FileTreeItem *)info->item;

	switch(type)
	{
		//次のファイル
		case ZIP_MOVE_NEXT:
			pi = _ITEM_NEXT(pi);
			break;
		//最初の子
		case ZIP_MOVE_FIRST_CHILD:
			pi = _ITEM(pi->i.first);
			break;
		//親
		case ZIP_MOVE_PARENT:
			pi = _ITEM(pi->i.parent);
			break;
		//トップ
		case ZIP_MOVE_TOP:
			pi = _ITEM(g_zip->tree.top);
			break;
		default:
			return 0;
	}

	if(!pi) return 0;

	info->item = pi;

	_set_zipfileinfo(info, pi);

	return 1;
}

/** ファイル内容をデコード
 *
 * バッファを確保して *ppbuf にセット。mFree() で解放する。 */

int zip_decode(ZipFileInfo *info,uint8_t **ppbuf)
{
	FileTreeItem *pi;
	uint8_t *outbuf;
	int ret;

	pi = (FileTreeItem *)info->item;

	//バッファ確保

	outbuf = (uint8_t *)mMalloc(pi->uncompsize, FALSE);
	if(!outbuf) return APPERR_MEMORY;

	//ファイル位置

	fsetpos(g_zip->fp, &pi->offset);

	//デコード

	if(pi->method == _METHOD_STORED)
		ret = _decode_stored(pi, outbuf);
	else
		ret = _decode_deflate(pi, outbuf);

	//

	if(ret == APPERR_OK)
		*ppbuf = outbuf;
	else
		mFree(outbuf);

	return ret;
}
